/*
 * Generic X2C block services.
 *
 * $LastChangedRevision: 480 $
 *
 * Copyright (c) 2013, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 */
#include "Services.h"
#include "TableStruct.h"
#include "BlockServicesX2C.h"

extern volatile tTableStruct *TableStruct;

void addBlockServices(tProtocol* protocol)
{
    tSERVICEFunction* pServiceTable = (tSERVICEFunction*) protocol->pServiceTable;

    pServiceTable[SV_ID_SVGETRAMBLOCK] = getRAMBlock;
    pServiceTable[SV_ID_GETBLOCKDATA] = getBlockData;
    pServiceTable[SV_ID_PUTBLOCKDATA] = putBlockData;
}

/* get data memory content */

/* this service is protected by payload length protection */
void getRAMBlock(tProtocol* protocol)
{
    void *addr;
    uint16 i;
    uint8 size, type;
    uint8 errorID = ERRORSuccess;
#if defined(__ADDRESS_WIDTH_8BIT__)
    const uint8 addrOffset = (uint8)2;
#elif defined(__ADDRESS_WIDTH_16BIT__)
    const uint8 addrOffset = (uint8)3;
#elif defined(__ADDRESS_WIDTH_32BIT__)
    const uint8 addrOffset = (uint8)5;
#else
#error ADDRESS WIDTH NOT DEFINED
#endif

    size = protocol->ucFRAMEData[addrOffset];
    type = protocol->ucFRAMEData[addrOffset + 1];

    if (((uint16) size + 2) > protocol->ucMaxCommSize)
    {
        errorID = ERRORSizeTooLarge;
    }
    else
    {
#if defined(__ADDRESS_WIDTH_8BIT__)
    	addr = (void*) ((uint8) protocol->ucFRAMEData[1];
#elif defined(__ADDRESS_WIDTH_16BIT__)
        addr = (void*) ((uint16) protocol->ucFRAMEData[1] + \
                ((uint16) protocol->ucFRAMEData[2] << 8));
#elif defined(__ADDRESS_WIDTH_32BIT__)
        addr = (void*) ((uint32) protocol->ucFRAMEData[1] + \
                ((uint32) protocol->ucFRAMEData[2] << 8) + \
                ((uint32) protocol->ucFRAMEData[3] << 16) + \
                ((uint32) protocol->ucFRAMEData[4] << 24));
#else
#error ADDRESS WIDTH NOT DEFINED
#endif

        switch (type)
        {
            case DATATYPE_8BIT:
            {
                uint8* addr8 = (uint8*) addr;
                for (i = 0; i < size; i++)
                {
                    protocol->ucFRAMEData[2 + i] = (uint8) *addr8++;
                }
                break;
            }
            case DATATYPE_16BIT:
            {
                uint16* addr16 = (uint16*) addr;
                for (i = 0; i < size; i += 2)
                {
                    protocol->ucFRAMEData[2 + i] = (uint8) *addr16;
                    protocol->ucFRAMEData[2 + i + 1] = (uint8) ((*addr16++) >> 8);
                }
                break;
            }
            case DATATYPE_32BIT:
            {
                uint32* addr32 = (uint32*) addr;
                for (i = 0; i < size; i += 4)
                {
                    protocol->ucFRAMEData[2 + i] = (uint8) *addr32;
                    protocol->ucFRAMEData[2 + i + 1] = (uint8) ((*addr32) >> 8);
                    protocol->ucFRAMEData[2 + i + 2] = (uint8) ((*addr32) >> 16);
                    protocol->ucFRAMEData[2 + i + 3] = (uint8) ((*addr32++) >> 24);
                }
                break;
            }
            default:
                errorID = ERRORFormat;
                break;
        }
    }

    if (errorID == ERRORSuccess)
    {
        protocol->ucFRAMESize = size + 2;
        protocol->ucFRAMEData[1] = errorID;
        protocol->pSnd_Enable((struct tProtocol*) protocol);
    }
    else
    {
        sendError(protocol, errorID);
    }
}

/* Downloads block parameters by using block address & block save function */
/* this service is NOT protected by payload length protection */

/* Reason: not required, response frame length <= request frame */
void putBlockData(tProtocol* protocol)
{
    uint16 id, i;
    uint16* addr;

#if defined(__ADDRESS_WIDTH_8BIT__)
    const uint16 addrOffset = (uint8)2;
    addr = (uint16*) ((uint8) protocol->ucFRAMEData[1];
#elif defined(__ADDRESS_WIDTH_16BIT__)
    const uint16 addrOffset = (uint8)3;
    addr = (uint16*) ((uint16) protocol->ucFRAMEData[1] + \
    		((uint16) protocol->ucFRAMEData[2] << 8));
#elif defined(__ADDRESS_WIDTH_32BIT__)
    const uint16 addrOffset = (uint8)5;
    addr = (uint16*) ((uint32) protocol->ucFRAMEData[1] + \
    		((uint32) protocol->ucFRAMEData[2] << 8) + \
    		((uint32) protocol->ucFRAMEData[3] << 16) + \
    		((uint32) protocol->ucFRAMEData[4] << 24));
#else
#error ADDRESS WIDTH NOT DEFINED
#endif
    id = *addr;

    i = 0;
    while ((id != TableStruct->TFncTable[i].iBlockID) && (TableStruct->TFncTable[i].iBlockID != 0))
    {
        i++;
    }
    if (TableStruct->TFncTable[i].iBlockID == 0)
    {
        /* if correct block id was not found -> return wrong block id error */
        sendError(protocol, ERRORBlkID);
    } else {
        if (TableStruct->TFncTable[i].pFSave(addr, protocol->ucFRAMEData + addrOffset,  \
                protocol->ucFRAMESize - addrOffset))
        {
            /* wrong data format */
            sendError(protocol, ERRORFormat);
        }
        else
        {
            protocol->ucFRAMESize = 2;
            protocol->ucFRAMEData[1] = ERRORSuccess;
            protocol->pSnd_Enable((struct tProtocol*) protocol);
        }
    }
}

/* Uploads block parameters by using block address & block load function */
/* this service is NOT protected by payload length protection */

/* Reason: currently not possible, see line w/ pFLoad instruction */
void getBlockData(tProtocol* protocol)
{
    uint16 id, i;
    uint16 *addr;

#if defined(__ADDRESS_WIDTH_8BIT__)
    	addr = (uint16*) ((uint8) protocol->ucFRAMEData[1];
#elif defined(__ADDRESS_WIDTH_16BIT__)
        addr = (uint16*) ((uint16) protocol->ucFRAMEData[1] + \
                ((uint16) protocol->ucFRAMEData[2] << 8));
#elif defined(__ADDRESS_WIDTH_32BIT__)
        addr = (uint16*) ((uint32) protocol->ucFRAMEData[1] + \
                ((uint32) protocol->ucFRAMEData[2] << 8) + \
                ((uint32) protocol->ucFRAMEData[3] << 16) + \
                ((uint32) protocol->ucFRAMEData[4] << 24));
#else
#error ADDRESS WIDTH NOT DEFINED
#endif
    id = *addr;

    i = 0;
    while ((id != TableStruct->TFncTable[i].iBlockID) && (TableStruct->TFncTable[i].iBlockID != 0))
    {
        i++;
    }
    if (TableStruct->TFncTable[i].iBlockID == 0)
    {
        /* if correct block id was not found -> return wrong block id error */
        sendError(protocol, ERRORBlkID);
    }
    else
    {
        protocol->ucFRAMESize = 2 + TableStruct->TFncTable[i].pFLoad(addr, protocol->ucFRAMEData + 2);
        protocol->ucFRAMEData[1] = ERRORSuccess;
        protocol->pSnd_Enable((struct tProtocol*) protocol);
    }
}
